/*********************************************************************/
/*                                                                   */
/*    chMonitor.c                                                    */
/*                                                                   */
/*    Read channel status/parameters                                 */ 
/*    Return info and OK or any failure message                      */
/*                                                                   */
/*    From HSCAENETLib-1-4-extended/sy546ctrl.c                      */
/*    Luca Boni, 15 settembre 2003                                   */
/*                                                                   */
/*    This program is free software: you can redistribute it	     */
/*    and/or modify it under the terms of the GNU General Public     */
/*    License version 2 as published by the Free Software Foundation */
/*                                                                   */
/*********************************************************************/

#include  <stdio.h>
#include  <string.h>
#include  <stdlib.h>
#include  <errno.h>
#include  <sys/types.h>
#include  <unistd.h>

#include  "HSCAENETLib.h"

/*******************************************************/
/*            internal & external constants            */     
/*******************************************************/

// Some of the Caenet Codes [word type]: -> see TECHNICAL INFORMATION MANUAL MOD.SY 546,
// MASTER-TO-SLAVE DATA PACKET DESCRIPTION pag.60
#define   READ_BOARDS_INFO                       0x3

/****************************/
/*    STRUCTURE OF DATA     */
/****************************/

//  The following structure contains all the useful information
//  about the characteristics of every board
struct  infoBoard {
  char    dummy[0];
  short   currUnit;
  short   vMax;
  short   iMax;
  char    reserved[40];
  short   rampMin;
  short   vRes;
  short   iRes;
  short   vDec;
  short   iDec;
  short   polarity;
  short   present;  
} boards[8];  

struct infoCH {
  // Status
  char    name[0];
  float   vMon;  // float because of 2 
  float   iMon;  // digit after comma
  char    *present;
  char    *pw;
  char    *tripping; 
  char    *status;
  char    *ramping;
  char    *SvMax;   
  // Parameters
  short   vSet;  // no digit after 
  short   iSet;  // comma (in any case)
  short   vMax;
  short   rUp;
  short   rDwn;
  short   trip;
} channels[96];

struct stats {
  int    On      ; 
  int    Off     ;
  int    Tripped ;
  int    UvV     ;
  int    OvV     ;
  int    OvC     ;
} statCrates[9];

// Globals
int       NUMBOARDS = 8, NUMCHANNELS = 12, bd;
float     pow_10[] = { 1.0, 10.0, 100.0, 1000.0 };
char      *currUMis[] = {"A","mA","uA","nA"};

static int	response, HSCAENETDev, code, crateNum;

/*******************************************************/
/*                  functions                          */ 
/*******************************************************/

void initCard (void);
void endCard (void);
static char *DecodeCAENETResp (int resp);
static void swap (unsigned char *a, unsigned char *b);
static int size_of_board (void);
static int buildBdInfo (struct infoBoard *board, unsigned char *cnetbuff, int shift);
static int getCrateInfo (int crateNum);


int main(int argc, char *argv[]) {

  float               scaleV, scaleI, iM;
  int                 ch, chnl, slotInfo;
  unsigned short      value1[4];  // Vmon<31..16> Vmon<15..0> Imon<15..0> Status<15..0>
                                  // strange: only Vmon<15..0> is my value
  unsigned short      value2[14];

  if (argc != 4) {
    printf("\nUsage:  chMonitor crateNum[1-8] board[0-7] log_slotInfo[0-1]\n\n");
    exit(-1);
  }

  initCard();
  
  crateNum = atoi(argv[1]);
  bd = atoi(argv[2]);
  slotInfo =  atoi(argv[3]);

  if (getCrateInfo(crateNum) != TUTTOK) {  // Gotta info about only Crate Configuration 
     printf("getCrateInfo error. %s\n", DecodeCAENETResp(response));
     endCard();
     exit(-1);
  }
  // go on only if board is present
  if (boards[bd].present & 1) {
  // now getting info about channels... 
    scaleV = pow_10[boards[bd].vDec];
    scaleI = pow_10[boards[bd].iDec];
    for (ch=0; ch<NUMCHANNELS; ch++) {
           chnl = bd*12 + ch;
           // about status and...
           code = chnl*256 + 1;  // %n01
           if ((response=HSCAENETComm(HSCAENETDev, code, crateNum, (NULL), (0), (value1))) != TUTTOK) {
              printf("HSCAENETComm status monitor failed. %s\n", DecodeCAENETResp(response));
              endCard();
	      exit(-1);              
           } else { 
   	      channels[chnl].vMon = value1[1]/scaleV;
              channels[chnl].iMon = value1[2]/scaleI;
              if (value1[3] & (1<<0))
	         channels[chnl].present = "Present";
              else  channels[chnl].present = "NotPresent";
              if (value1[3] & (1<<8)) 
	         channels[chnl].SvMax = "Vmax"; 
              else channels[chnl].SvMax = "-"; 
              if (value1[3] & (1<<9)) { 
	         channels[chnl].tripping = "Trip"; 
	         statCrates[crateNum].Tripped ++;
              }
              else channels[chnl].tripping = "NO";
              channels[chnl].status = "OK";  
              if (value1[3] & (1<<10)) {              
	         channels[chnl].status = "OvV";
                 statCrates[crateNum].OvV ++;
              }
              if (value1[3] & (1<<11)) {
	         channels[chnl].status = "UvV";
                 statCrates[crateNum].UvV ++;
              }    
              if (value1[3] & (1<<12)) {
	         channels[chnl].status = "OvC";
                 statCrates[crateNum].OvC ++;
              }                           
              if (value1[3] & (1<<13)) 
	         channels[chnl].ramping = "Down";
              else if (value1[3] & (1<<14))
	         channels[chnl].ramping = "Up";
              else channels[chnl].ramping = "-";
              if (value1[3] & (1<<15)) {
	         channels[chnl].pw = "On";
                 statCrates[crateNum].On ++;
              } else {
                 channels[chnl].pw = "Off";
                 statCrates[crateNum].Off ++;
              }
              // ... about parameters:
              code = chnl*256 + 2;  // %n02
              if ((response=HSCAENETComm(HSCAENETDev, code, crateNum, (NULL), (0), (value2))) != TUTTOK) {
                 printf("HSCAENETComm values monitor failed. %s\n", DecodeCAENETResp(response));
		 endCard();
		 exit(-1);
              } else {
                 channels[chnl].vSet = (int) (value2[7]/scaleV);
                 channels[chnl].iSet = (int) (value2[8]/scaleI);
                 channels[chnl].vMax = value2[9];
                 channels[chnl].rUp  = value2[10]; 
                 channels[chnl].rDwn = value2[11];
                 channels[chnl].trip = value2[12]/10;
             }  // end if
           }  // end if           
    }  // end for

   if (slotInfo == 1) {
	printf ("\nSlot %d: ", bd);
	if (boards[bd].present & 1) {
    	    printf ("Board %dCH  ");
    	    printf( (boards[bd].polarity & 1) ? "POSITIVE" : "NEGATIVE");
    	    printf("  %dV",boards[bd].vMax);
    	    iM = (float)boards[bd].iMax / pow_10[boards[bd].iDec];     
    	    printf(" %8.2f",iM); 
    	    printf("%s\n",currUMis[(int)boards[bd].currUnit]);
    	    printf("        resolution: %dV Vset/mon &",boards[bd].vRes);  
    	    printf(" %d%s Iset/mon,",boards[bd].iRes,currUMis[(int)boards[bd].currUnit]);
    	    printf(" ramp-Up/Down: %dV min\n",boards[bd].rampMin);
    	    printf("        significant figures: %d Vset/mon &",boards[bd].vDec);
    	    printf(" %d Iset/mon\n\n",boards[bd].iDec);
	} else printf ("Not Present\n\n");
   }

   for (ch=0; ch<NUMCHANNELS; ch++) {
      chnl = bd*12 + ch;
      printf ("%d:\t%.2fV\t%.2f%s\t%s\t%s\t%s\t%s\t%s\t%s", 
              ch, 
              channels[chnl].vMon, 
              channels[chnl].iMon, currUMis[(int)boards[bd].currUnit], 
              channels[chnl].present, 
              channels[chnl].pw, 
              channels[chnl].tripping, 
              channels[chnl].status, 
              channels[chnl].ramping, 
              channels[chnl].SvMax
	     );
      printf ("\t%dV\t%d%s\t%dV\t%dV/s\t%dV/s\t%d\n",
              channels[chnl].vSet, 
	      channels[chnl].iSet, currUMis[(int)boards[bd].currUnit],
              channels[chnl].vMax,
              channels[chnl].rUp,
              channels[chnl].rDwn,
              channels[chnl].trip
             );
   } // end for

  // board not present .. fill log with zeros
  } else {
    for (ch=0; ch<NUMCHANNELS; ch++) {
         printf ("%d:\t0.00V\t0.00nA\tNotPresent\t-\t-\t-\t-\t-",ch);
	 printf ("\t0V\t0nA\t0V\t0V/s\t0V/s\t0\n");
    }
   }

   // statistics of all channels/boards of a crate
   printf("on %d, off %d, tripped %d, UvV %d, OvV %d, OvC %d\n",
	  statCrates[crateNum].On, statCrates[crateNum].Off, statCrates[crateNum].Tripped,
          statCrates[crateNum].UvV, statCrates[crateNum].OvV, statCrates[crateNum].OvC);

  puts("OK");
  endCard();
  return EXIT_SUCCESS;

}

/*******************************************************/
/*                   Subroutine initCard               */
/*  : connecting A1303                                 */  
/*******************************************************/

void initCard (void) {  
  
  int    a1303index = 0;
  long   timeout = 100;  // that is 1 sec

  HSCAENETDev = HSCAENETCardInit("A1303", &a1303index);
  if ( HSCAENETDev == -1 ) {
     puts("HSCAENETCardInit failed");
     exit(-1);
  } 
  if ( ( response = HSCAENETCardReset(HSCAENETDev) ) != TUTTOK ) {
        printf("HSCAENETCardReset failed. %s\n", DecodeCAENETResp(response));
	endCard();
	exit(-1);
  }
  if ( ( response = HSCAENETTimeout(HSCAENETDev, timeout) ) != TUTTOK ) {
        printf("HSCAENETTimeout failed. %s\n", DecodeCAENETResp(response));
	endCard();
	exit(-1);
  }

}

/*******************************************************/
/*                   Subroutine endCard                */
/*  : disconnecting A1303                              */
/*******************************************************/

void endCard (void) { HSCAENETCardEnd(HSCAENETDev); }

/*******************************************************/
/*             Subroutine DecodeCAENETResp             */
/*  : decode response from CAENET                      */     
/*  : return error response                            */
/*******************************************************/
static char *DecodeCAENETResp (int resp) {

  static char buf[80];

  if ( resp == E_OS )
     strcpy(buf, strerror(errno));
  else
     switch( resp ) {
        case E_LESS_DATA:
           strcpy(buf, "CaeNet Error: Less data received" );
           break;
        case E_TIMEOUT:
           strcpy(buf, "CaeNet Error: Timeout expired" );
           break;
        case E_NO_DEVICE:
           strcpy(buf, "CaeNet Error: High Speed CAENET device not found" );
           break;
        case E_A303_BUSY:
           sprintf(buf, "Another instance of a303lib.dll already loaded" );
           break;
        default:
           sprintf(buf, "CaeNet Error: %d", resp );
           break;
     }

  return buf;
}

/*******************************************************/
/*** BEGIN useful Utilities ***/

static void swap (unsigned char *a, unsigned char *b) {
  unsigned char temp;
  temp = *a;
  *a = *b;
  *b = temp;
}

static void swap_byte (unsigned char *buff, int size) {
  int  i;
  for ( i=0 ; i<size ; i += 2 )
     swap(buff+i,buff+i+1);
}

static int size_of_board (void) {  
  struct infoBoard b;
  return sizeof(b.dummy) + sizeof(b.currUnit) + sizeof(b.vMax)     + sizeof(b.iMax) 
    + sizeof(b.reserved) + sizeof(b.rampMin)  + sizeof(b.vRes)     + sizeof(b.iRes)
    + sizeof(b.vDec)     + sizeof(b.iDec)     + sizeof(b.polarity) + sizeof(b.present);
}

/* at service of following subroutine */
static int buildBdInfo (struct infoBoard *board, unsigned char *cnetbuff, int shift) {

  int i = shift; 

  i += sizeof(board->dummy);

  memcpy((char *)&board->currUnit, cnetbuff+i, sizeof(board->currUnit));
  swap_byte((char *)&board->currUnit,sizeof(board->currUnit));
  i += sizeof(board->currUnit); 

  memcpy((char *)&board->vMax, cnetbuff+i, sizeof(board->vMax));
  swap_byte((char *)&board->vMax,sizeof(board->vMax));
  i += sizeof(board->vMax);

  memcpy((char *)&board->iMax, cnetbuff+i, sizeof(board->iMax));
  swap_byte((char *)&board->iMax,sizeof(board->iMax));
  i += sizeof(board->iMax) + sizeof(board->reserved); 

  memcpy((char *)&board->rampMin, cnetbuff+i, sizeof(board->rampMin));
  swap_byte((char *)&board->rampMin,sizeof(board->rampMin));
  i += sizeof(board->rampMin); 

  memcpy((char *)&board->vRes, cnetbuff+i, sizeof(board->vRes)); 
  swap_byte((char *)&board->vRes,sizeof(board->vRes));
  i += sizeof(board->vRes); 

  memcpy((char *)&board->iRes, cnetbuff+i, sizeof(board->iRes));
  swap_byte((char *)&board->iRes,sizeof(board->iRes));
  i += sizeof(board->iRes); 

  memcpy((char *)&board->vDec, cnetbuff+i, sizeof(board->vDec));
  swap_byte((char *)&board->vDec,sizeof(board->vDec));
  i += sizeof(board->vDec); 

  memcpy((char *)&board->iDec, cnetbuff+i, sizeof(board->iDec));
  swap_byte((char *)&board->iDec,sizeof(board->iDec));
  i += sizeof(board->iDec); 

  memcpy((char *)&board->polarity, cnetbuff+i, sizeof(board->polarity));
  swap_byte((char *)&board->polarity,sizeof(board->polarity));
  i += sizeof(board->polarity); 

  memcpy((char *)&board->present, cnetbuff+i, sizeof(board->present));
  swap_byte((char *)&board->present,sizeof(board->present));
  i += sizeof(board->present); 

  return TUTTOK;
}

/* building of boards[] */
static int getCrateInfo (int crateNum) { 

  int 		 nBd;
  unsigned char  cnetbuff[MAX_LENGTH_FIFO];  // sizeof -> 4096

  code = READ_BOARDS_INFO;
  if ((response=HSCAENETComm(HSCAENETDev, code, crateNum, (NULL), (0), (cnetbuff))) != TUTTOK) {
     printf("HSCAENETComm Crate info getting failed. %s\n", DecodeCAENETResp(response));
     endCard();
     return response;
  } else {
     swap_byte(cnetbuff, NUMBOARDS*size_of_board());
     for (nBd=0; nBd<NUMBOARDS; nBd++)
        buildBdInfo((struct infoBoard *)&boards[nBd],cnetbuff,nBd*size_of_board());
  }  

  return TUTTOK;
}
