/*********************************************************************/
/*                                                                   */
/*    setChannels.c                                                  */
/*                                                                   */
/*    Set parameters for all channels   	  		     */
/*    Return OK or any failure messages				     */
/*								     */
/*    NOTE:							     */
/*    Only Iset and Trip time must be setted up for each channel     */
/*    Other parameters will be applied for all channels all together */ 
/*                                                                   */
/*    From HSCAENETLib-1-4/sy546ctrl.c                               */
/*    Luca Boni, 16 settembre 2003                                   */
/*                                                                   */
/*    This program is free software: you can redistribute it	     */
/*    and/or modify it under the terms of the GNU General Public     */
/*    License version 2 as published by the Free Software Foundation */
/*                                                                   */
/*********************************************************************/

#include  <stdio.h>
#include  <string.h>
#include  <stdlib.h>
#include  <errno.h>
#include  <sys/types.h>
#include  <unistd.h>

#include  "HSCAENETLib.h"

/*******************************************************/
/*            internal & external constants            */     
/*******************************************************/

// Some of the Caenet Codes [word type]: -> see TECHNICAL INFORMATION MANUAL MOD.SY 546,
// MASTER-TO-SLAVE DATA PACKET DESCRIPTION pag.60
#define   READ_BOARDS_INFO                       0x3

/****************************/
/*    STRUCTURE OF DATA     */
/****************************/

//  The following structure contains all the useful information
//  about the characteristics of every board
struct  infoBoard {
  char    dummy[0];
  short   currUnit;
  short   vMax;
  short   iMax;
  char    reserved[40];
  short   rampMin;
  short   vRes;
  short   iRes;
  short   vDec;
  short   iDec;
  short   polarity;
  short   present;  
} boards[8];  

struct infoCH {
  // Status
  char    name[0];
  float   vMon;  // float because of 2 
  float   iMon;  // digit after comma
  char    *present;
  char    *pw;
  char    *tripping; 
  char    *status;
  char    *ramping;
  char    *SvMax;   
  // Parameters
  short   vSet;  // no digit after 
  short   iSet;  // comma (in any case)
  short   vMax;
  short   rUp;
  short   rDwn;
  short   trip;
} channels[96];

struct stats {
  int    On      ; 
  int    Off     ;
  int    Tripped ;
  int    UvV     ;
  int    OvV     ;
  int    OvC     ;
} statCrates[9];

// Globals
int       NUMBOARDS = 8, NUMCHANNELS = 12;
float     pow_10[] = { 1.0, 10.0, 100.0, 1000.0 };

static int	response, HSCAENETDev, code;

/*******************************************************/
/*                  functions                          */ 
/*******************************************************/

void initCard (void);
void endCard (void);
static char *DecodeCAENETResp (int resp);
static void swap (unsigned char *a, unsigned char *b);
static int size_of_board (void);
static int buildBdInfo (struct infoBoard *board, unsigned char *cnetbuff, int shift);
static int getCrateInfo (int crateNum);


int main(int argc, char *argv[]) {

  int         		crateNum, bd, channel, ch;
  float       		input_value, scale;
  unsigned short	cnet_buff[1], flagMask[1];


  if (argc != 9) {
    printf("\nUsage:  setChannels crateNum[1-8] board[0-7] VSET[V] ISET[nA] VMAX[V] RUP[V/s] RDWN[V/s] TRIP[10th of sec]\n\n");
    exit(-1);
  }

  initCard();
  
  crateNum =  atoi(argv[1]);
  bd = atoi(argv[2]);

  if (getCrateInfo(crateNum) != TUTTOK)  { // Get information about the Crate Configuration 
     printf("getCrateInfo error. %s\n", DecodeCAENETResp(response));
     exit(-1);
  }

  if (boards[bd].present & 1) {  // execute only if board exists
  /* NOTE:
     sleep() is needed only for those parameters
     that will be setted up for all channels all together,
     because is needed a bit more time that apply Iset
     and Trip time that are applied singularly.
     This is only an optimization tip.
  */
  channel = bd * NUMCHANNELS;  // all channels
  // VSET
  code = channel*256 + 16;  // %n10
  scale = pow_10[boards[bd].vDec];
  input_value = atof(argv[3]);
  input_value *= scale;
  cnet_buff[0] = (unsigned short)input_value;
  if ((response=HSCAENETComm(HSCAENETDev, code, crateNum, cnet_buff, sizeof(cnet_buff), (NULL))) != TUTTOK) {
	printf("HSCAENETComm: error setting parameter VSET. %s\n", DecodeCAENETResp(response));
	endCard();
	exit(-1);
  }
  sleep (1);
  // ISET
  for (ch=0; ch<NUMCHANNELS; ch++) {   // otherwise not equal for each channel
    code = (channel + ch) * 256 + 18;  // %n12
    scale = pow_10[boards[bd].iDec];
    input_value = atof(argv[4]);
    input_value *= scale;
    cnet_buff[0] = (unsigned short)input_value;
    if ((response=HSCAENETComm(HSCAENETDev, code, crateNum, cnet_buff, sizeof(cnet_buff), (NULL))) != TUTTOK) {
	printf("HSCAENETComm: error setting parameter ISET. %s\n", DecodeCAENETResp(response));
	endCard();
	exit(-1);
    }
  }
  // VMAX
  code = channel*256 + 20;  // %n14
  input_value = atof(argv[5]);
  cnet_buff[0] = (unsigned short)input_value;
  if ((response=HSCAENETComm(HSCAENETDev, code, crateNum, cnet_buff, sizeof(cnet_buff), (NULL))) != TUTTOK) {
	printf("HSCAENETComm: error setting parameter VMAX. %s\n", DecodeCAENETResp(response));
	endCard();
	exit(-1);
  }
  sleep (1);
  // RUP
  code = channel*256 + 21;  // %n15
  input_value = atof(argv[6]);
  cnet_buff[0] = (unsigned short)input_value;
  if ((response=HSCAENETComm(HSCAENETDev, code, crateNum, cnet_buff, sizeof(cnet_buff), (NULL))) != TUTTOK) {
	printf("HSCAENETComm: error setting parameter RUP. %s\n", DecodeCAENETResp(response));
	endCard();
	exit(-1);
  }
  sleep (1);
  // RDWN
  code = channel*256 + 22;  // %n16
  input_value = atof(argv[7]);
  cnet_buff[0] = (unsigned short)input_value;
  if ((response=HSCAENETComm(HSCAENETDev, code, crateNum, cnet_buff, sizeof(cnet_buff), (NULL))) != TUTTOK) {
	printf("HSCAENETComm: error setting parameter RDWN. %s\n", DecodeCAENETResp(response));
	endCard();
	exit(-1);
  }
  sleep (1);
  // TRIP
  for (ch=0; ch<NUMCHANNELS; ch++) {  // otherwise not equal for each channel
    code = (channel +ch) * 256 + 23;  // %n17
    input_value = atof(argv[8]);
    input_value *= 10;  // Trip is in 10-th of sec
    cnet_buff[0] = (unsigned short)input_value;
    if ((response=HSCAENETComm(HSCAENETDev, code, crateNum, cnet_buff, sizeof(cnet_buff), (NULL))) != TUTTOK) {
	printf("HSCAENETComm: error setting parameter TRIP. %s\n", DecodeCAENETResp(response));
	endCard();
	exit(-1);
    }
  }

  flagMask[0] = 55496;  // now H.V.=1 PWON=1 ON/OFF=1: b15..b0 is 1101100011001000
  code = channel*256 + 24;  // %n18 
  if ((response=HSCAENETComm(HSCAENETDev, code, crateNum, flagMask, sizeof(flagMask), (NULL))) != TUTTOK) {
       printf("HSCAENETComm switch On failed. %s\n", DecodeCAENETResp(response));
       endCard();
       exit(-1);
  }

  } else puts("Error, board not present!");

  puts("OK");
  endCard();
  return EXIT_SUCCESS;

}

/*******************************************************/
/*                   Subroutine initCard               */
/*  : connecting A1303                                 */  
/*******************************************************/

void initCard (void) {  
  
  int    a1303index = 0;
  long   timeout = 100;  // that is 1 sec

  HSCAENETDev = HSCAENETCardInit("A1303", &a1303index);
  if ( HSCAENETDev == -1 ) {
     puts("HSCAENETCardInit failed");
     exit(-1);
  } 
  if ( ( response = HSCAENETCardReset(HSCAENETDev) ) != TUTTOK ) {
        printf("HSCAENETCardReset failed. %s\n", DecodeCAENETResp(response));
	endCard();
	exit(-1);
  }
  if ( ( response = HSCAENETTimeout(HSCAENETDev, timeout) ) != TUTTOK ) {
        printf("HSCAENETTimeout failed. %s\n", DecodeCAENETResp(response));
	endCard();
	exit(-1);
  }

}

/*******************************************************/
/*                   Subroutine endCard                */
/*  : disconnecting A1303                              */
/*******************************************************/

void endCard (void) { HSCAENETCardEnd(HSCAENETDev); }

/*******************************************************/
/*             Subroutine DecodeCAENETResp             */
/*  : decode response from CAENET                      */     
/*  : return error response                            */
/*******************************************************/
static char *DecodeCAENETResp (int resp) {

  static char buf[80];

  if ( resp == E_OS )
     strcpy(buf, strerror(errno));
  else
     switch( resp ) {
        case E_LESS_DATA:
           strcpy(buf, "CaeNet Error: Less data received" );
           break;
        case E_TIMEOUT:
           strcpy(buf, "CaeNet Error: Timeout expired" );
           break;
        case E_NO_DEVICE:
           strcpy(buf, "CaeNet Error: High Speed CAENET device not found" );
           break;
        case E_A303_BUSY:
           sprintf(buf, "Another instance of a303lib.dll already loaded" );
           break;
        default:
           sprintf(buf, "CaeNet Error: %d", resp );
           break;
     }

  return buf;
}

/*******************************************************/
/*** BEGIN useful Utilities ***/

static void swap (unsigned char *a, unsigned char *b) {
  unsigned char temp;
  temp = *a;
  *a = *b;
  *b = temp;
}

static void swap_byte (unsigned char *buff, int size) {
  int  i;
  for ( i=0 ; i<size ; i += 2 )
     swap(buff+i,buff+i+1);
}

static int size_of_board (void) {  
  struct infoBoard b;
  return sizeof(b.dummy) + sizeof(b.currUnit) + sizeof(b.vMax)     + sizeof(b.iMax) 
    + sizeof(b.reserved) + sizeof(b.rampMin)  + sizeof(b.vRes)     + sizeof(b.iRes)
    + sizeof(b.vDec)     + sizeof(b.iDec)     + sizeof(b.polarity) + sizeof(b.present);
}

/* at service of following subroutine */
static int buildBdInfo (struct infoBoard *board, unsigned char *cnetbuff, int shift) {

  int i = shift; 

  i += sizeof(board->dummy);

  memcpy((char *)&board->currUnit, cnetbuff+i, sizeof(board->currUnit));
  swap_byte((char *)&board->currUnit,sizeof(board->currUnit));
  i += sizeof(board->currUnit); 

  memcpy((char *)&board->vMax, cnetbuff+i, sizeof(board->vMax));
  swap_byte((char *)&board->vMax,sizeof(board->vMax));
  i += sizeof(board->vMax);

  memcpy((char *)&board->iMax, cnetbuff+i, sizeof(board->iMax));
  swap_byte((char *)&board->iMax,sizeof(board->iMax));
  i += sizeof(board->iMax) + sizeof(board->reserved); 

  memcpy((char *)&board->rampMin, cnetbuff+i, sizeof(board->rampMin));
  swap_byte((char *)&board->rampMin,sizeof(board->rampMin));
  i += sizeof(board->rampMin); 

  memcpy((char *)&board->vRes, cnetbuff+i, sizeof(board->vRes)); 
  swap_byte((char *)&board->vRes,sizeof(board->vRes));
  i += sizeof(board->vRes); 

  memcpy((char *)&board->iRes, cnetbuff+i, sizeof(board->iRes));
  swap_byte((char *)&board->iRes,sizeof(board->iRes));
  i += sizeof(board->iRes); 

  memcpy((char *)&board->vDec, cnetbuff+i, sizeof(board->vDec));
  swap_byte((char *)&board->vDec,sizeof(board->vDec));
  i += sizeof(board->vDec); 

  memcpy((char *)&board->iDec, cnetbuff+i, sizeof(board->iDec));
  swap_byte((char *)&board->iDec,sizeof(board->iDec));
  i += sizeof(board->iDec); 

  memcpy((char *)&board->polarity, cnetbuff+i, sizeof(board->polarity));
  swap_byte((char *)&board->polarity,sizeof(board->polarity));
  i += sizeof(board->polarity); 

  memcpy((char *)&board->present, cnetbuff+i, sizeof(board->present));
  swap_byte((char *)&board->present,sizeof(board->present));
  i += sizeof(board->present);
  
  return TUTTOK;

}

/* building of boards[] */
static int getCrateInfo (int crateNum) { 

  int            nBd;
  unsigned char  cnetbuff[MAX_LENGTH_FIFO];  // sizeof -> 4096

  code = READ_BOARDS_INFO;
  if ((response=HSCAENETComm(HSCAENETDev, code, crateNum, (NULL), (0), (cnetbuff))) != TUTTOK) {
     printf("HSCAENETComm Crate info getting failed. %s\n", DecodeCAENETResp(response));
     endCard();
     return response;
  } else {
     swap_byte(cnetbuff, NUMBOARDS*size_of_board());
     for (nBd=0; nBd<NUMBOARDS; nBd++) 
        buildBdInfo((struct infoBoard *)&boards[nBd],cnetbuff,nBd*size_of_board());
  }

  return TUTTOK;
}
